/*
 * block.c --- iterate over all blocks in an inode
 * 
 * Copyright (C) 1993, 1994, 1995, 1996 Theodore Ts'o.
 *
 * %Begin-Header%
 * This file may be redistributed under the terms of the GNU Public
 * License.
 * %End-Header%
 */

#include <stdio.h>
#include <string.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "ext2_fs.h"
#include "ext2fs.h"
#include "block.h"

#ifdef EXT_DEBUG
void show_inode(ext2_ino_t ino)
{
        printf("inode: %u\n", ino);
}
#else
#define show_inode(ino) do { } while (0)
#endif

static int block_iterate_ind(blk_t *ind_block, blk_t ref_block,
			     int ref_offset, struct block_context *ctx)
{
	int	ret = 0, changed = 0;
	int	i, flags, limit, offset;
	blk_t	*block_nr;

	limit = ctx->fs->blocksize >> 2;
	if (!(ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
	    !(ctx->flags & BLOCK_FLAG_DATA_ONLY))
		ret = (*ctx->func)(ctx->fs, ind_block,
				   BLOCK_COUNT_IND, ref_block,
				   ref_offset, ctx->priv_data);
	if (!*ind_block || (ret & BLOCK_ABORT)) {
		ctx->bcount += limit;
		return ret;
	}
	if (*ind_block >= ctx->fs->super->s_blocks_count ||
	    *ind_block < ctx->fs->super->s_first_data_block) {
		ctx->errcode = EXT2_ET_BAD_IND_BLOCK;
		ret |= BLOCK_ERROR;
		return ret;
	}
	ctx->errcode = ext2fs_read_ind_block(ctx->fs, *ind_block, 
					     ctx->ind_buf);
	if (ctx->errcode) {
		ret |= BLOCK_ERROR;
		return ret;
	}

	block_nr = (blk_t *) ctx->ind_buf;
	offset = 0;
	if (ctx->flags & BLOCK_FLAG_APPEND) {
		for (i = 0; i < limit; i++, ctx->bcount++, block_nr++) {
			flags = (*ctx->func)(ctx->fs, block_nr, ctx->bcount,
					     *ind_block, offset, 
					     ctx->priv_data);
			changed	|= flags;
			if (flags & BLOCK_ABORT) {
				ret |= BLOCK_ABORT;
				break;
			}
			offset += sizeof(blk_t);
		}
	} else {
		for (i = 0; i < limit; i++, ctx->bcount++, block_nr++) {
			if (*block_nr == 0)
				continue;
			flags = (*ctx->func)(ctx->fs, block_nr, ctx->bcount,
					     *ind_block, offset, 
					     ctx->priv_data);
			changed	|= flags;
			if (flags & BLOCK_ABORT) {
				ret |= BLOCK_ABORT;
				break;
			}
			offset += sizeof(blk_t);
		}
	}
	if (changed & BLOCK_CHANGED) {
		ctx->errcode = ext2fs_write_ind_block(ctx->fs, *ind_block,
						      ctx->ind_buf);
		if (ctx->errcode)
			ret |= BLOCK_ERROR | BLOCK_ABORT;
	}
	if ((ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
	    !(ctx->flags & BLOCK_FLAG_DATA_ONLY) &&
	    !(ret & BLOCK_ABORT))
		ret |= (*ctx->func)(ctx->fs, ind_block,
				    BLOCK_COUNT_IND, ref_block,
				    ref_offset, ctx->priv_data);
	return ret;
}
	
static int block_iterate_dind(blk_t *dind_block, blk_t ref_block,
			      int ref_offset, struct block_context *ctx)
{
	int	ret = 0, changed = 0;
	int	i, flags, limit, offset;
	blk_t	*block_nr;

	limit = ctx->fs->blocksize >> 2;
	if (!(ctx->flags & (BLOCK_FLAG_DEPTH_TRAVERSE |
			    BLOCK_FLAG_DATA_ONLY)))
		ret = (*ctx->func)(ctx->fs, dind_block,
				   BLOCK_COUNT_DIND, ref_block,
				   ref_offset, ctx->priv_data);
	if (!*dind_block || (ret & BLOCK_ABORT)) {
		ctx->bcount += limit*limit;
		return ret;
	}
	if (*dind_block >= ctx->fs->super->s_blocks_count ||
	    *dind_block < ctx->fs->super->s_first_data_block) {
		ctx->errcode = EXT2_ET_BAD_DIND_BLOCK;
		ret |= BLOCK_ERROR;
		return ret;
	}
	ctx->errcode = ext2fs_read_ind_block(ctx->fs, *dind_block, 
					     ctx->dind_buf);
	if (ctx->errcode) {
		ret |= BLOCK_ERROR;
		return ret;
	}

	block_nr = (blk_t *) ctx->dind_buf;
	offset = 0;
	if (ctx->flags & BLOCK_FLAG_APPEND) {
		for (i = 0; i < limit; i++, block_nr++) {
			flags = block_iterate_ind(block_nr,
						  *dind_block, offset,
						  ctx);
			changed |= flags;
			if (flags & (BLOCK_ABORT | BLOCK_ERROR)) {
				ret |= flags & (BLOCK_ABORT | BLOCK_ERROR);
				break;
			}
			offset += sizeof(blk_t);
		}
	} else {
		for (i = 0; i < limit; i++, block_nr++) {
			if (*block_nr == 0) {
				ctx->bcount += limit;
				continue;
			}
			flags = block_iterate_ind(block_nr,
						  *dind_block, offset,
						  ctx);
			changed |= flags;
			if (flags & (BLOCK_ABORT | BLOCK_ERROR)) {
				ret |= flags & (BLOCK_ABORT | BLOCK_ERROR);
				break;
			}
			offset += sizeof(blk_t);
		}
	}
	if (changed & BLOCK_CHANGED) {
		ctx->errcode = ext2fs_write_ind_block(ctx->fs, *dind_block,
						      ctx->dind_buf);
		if (ctx->errcode)
			ret |= BLOCK_ERROR | BLOCK_ABORT;
	}
	if ((ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
	    !(ctx->flags & BLOCK_FLAG_DATA_ONLY) &&
	    !(ret & BLOCK_ABORT))
		ret |= (*ctx->func)(ctx->fs, dind_block,
				    BLOCK_COUNT_DIND, ref_block,
				    ref_offset, ctx->priv_data);
	return ret;
}
	
static int block_iterate_tind(blk_t *tind_block, blk_t ref_block,
			      int ref_offset, struct block_context *ctx)
{
	int	ret = 0, changed = 0;
	int	i, flags, limit, offset;
	blk_t	*block_nr;

	limit = ctx->fs->blocksize >> 2;
	if (!(ctx->flags & (BLOCK_FLAG_DEPTH_TRAVERSE |
			    BLOCK_FLAG_DATA_ONLY)))
		ret = (*ctx->func)(ctx->fs, tind_block,
				   BLOCK_COUNT_TIND, ref_block,
				   ref_offset, ctx->priv_data);
	if (!*tind_block || (ret & BLOCK_ABORT)) {
		ctx->bcount += limit*limit*limit;
		return ret;
	}
	if (*tind_block >= ctx->fs->super->s_blocks_count ||
	    *tind_block < ctx->fs->super->s_first_data_block) {
		ctx->errcode = EXT2_ET_BAD_TIND_BLOCK;
		ret |= BLOCK_ERROR;
		return ret;
	}
	ctx->errcode = ext2fs_read_ind_block(ctx->fs, *tind_block, 
					     ctx->tind_buf);
	if (ctx->errcode) {
		ret |= BLOCK_ERROR;
		return ret;
	}

	block_nr = (blk_t *) ctx->tind_buf;
	offset = 0;
	if (ctx->flags & BLOCK_FLAG_APPEND) {
		for (i = 0; i < limit; i++, block_nr++) {
			flags = block_iterate_dind(block_nr,
						   *tind_block,
						   offset, ctx);
			changed |= flags;
			if (flags & (BLOCK_ABORT | BLOCK_ERROR)) {
				ret |= flags & (BLOCK_ABORT | BLOCK_ERROR);
				break;
			}
			offset += sizeof(blk_t);
		}
	} else {
		for (i = 0; i < limit; i++, block_nr++) {
			if (*block_nr == 0) {
				ctx->bcount += limit*limit;
				continue;
			}
			flags = block_iterate_dind(block_nr,
						   *tind_block,
						   offset, ctx);
			changed |= flags;
			if (flags & (BLOCK_ABORT | BLOCK_ERROR)) {
				ret |= flags & (BLOCK_ABORT | BLOCK_ERROR);
				break;
			}
			offset += sizeof(blk_t);
		}
	}
	if (changed & BLOCK_CHANGED) {
		ctx->errcode = ext2fs_write_ind_block(ctx->fs, *tind_block,
						      ctx->tind_buf);
		if (ctx->errcode)
			ret |= BLOCK_ERROR | BLOCK_ABORT;
	}
	if ((ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
	    !(ctx->flags & BLOCK_FLAG_DATA_ONLY) &&
	    !(ret & BLOCK_ABORT))
		ret |= (*ctx->func)(ctx->fs, tind_block,
				    BLOCK_COUNT_TIND, ref_block,
				    ref_offset, ctx->priv_data);
	
	return ret;
}
	
errcode_t ext2fs_block_iterate2(ext2_filsys fs,
				ext2_ino_t ino,
				int	flags,
				char *block_buf,
				int (*func)(ext2_filsys fs,
					    blk_t	*blocknr,
					    e2_blkcnt_t	blockcnt,
					    blk_t	ref_blk,
					    int		ref_offset,
					    void	*priv_data),
				void *priv_data)
{
	int	i;
	int	ret = 0;
	blk_t	blocks[EXT2_N_BLOCKS];	/* directory data blocks */
	struct ext2_inode inode;
	errcode_t	retval;
	struct block_context ctx;
	int	limit;
	struct ext3_extent_header *eh;

	EXT2_CHECK_MAGIC(fs, EXT2_ET_MAGIC_EXT2FS_FILSYS);

	ctx.errcode = ext2fs_read_inode(fs, ino, &inode);
	if (ctx.errcode)
		return ctx.errcode;

	/*
	 * Check to see if we need to limit large files
	 */
	if (flags & BLOCK_FLAG_NO_LARGE) {
		if (!LINUX_S_ISDIR(inode.i_mode) &&
		    (inode.i_size_high != 0))
			return EXT2_ET_FILE_TOO_BIG;
	}

	/* The in-memory inode may have been changed by e2fsck */
	retval = ext2fs_get_blocks(fs, ino, blocks);
	if (retval)
		return retval;

	limit = fs->blocksize >> 2;

	ctx.fs = fs;
	ctx.func = func;
	ctx.priv_data = priv_data;
	ctx.flags = flags;
	ctx.bcount = 0;
	if (block_buf) {
		ctx.ind_buf = block_buf;
	} else {
		retval = ext2fs_get_mem(fs->blocksize * 3, &ctx.ind_buf);
		if (retval)
			return retval;
	}
	ctx.dind_buf = ctx.ind_buf + fs->blocksize;
	ctx.tind_buf = ctx.dind_buf + fs->blocksize;

	/*
	 * Iterate over the HURD translator block (if present)
	 */
	if ((fs->super->s_creator_os == EXT2_OS_HURD) &&
	    !(flags & BLOCK_FLAG_DATA_ONLY)) {
		if (inode.osd1.hurd1.h_i_translator) {
			ret |= (*ctx.func)(fs,
					   &inode.osd1.hurd1.h_i_translator,
					   BLOCK_COUNT_TRANSLATOR,
					   0, 0, priv_data);
			if (ret & BLOCK_ABORT)
				goto abort_exit;
		}
	}

	/* Iterate over normal data blocks with extents.  We checked that
	 * EXT4_EXTENTS_FL and EXT3_EXT_MAGIC match in e2fsck_pass1(). */
	if (inode.i_flags & EXT4_EXTENTS_FL) {
		eh = (struct ext3_extent_header *)blocks;
		if (eh->eh_magic == EXT3_EXT_MAGIC) {
			show_inode(ino);
			ret |= block_iterate_extents(eh, 0, 0, &ctx);
		} else {
			ret |= BLOCK_ERROR;
		}
		goto abort_exit;
	}

	/*
	 * Iterate over normal data blocks
	 */
	for (i = 0; i < EXT2_NDIR_BLOCKS ; i++, ctx.bcount++) {
		if (blocks[i] || (flags & BLOCK_FLAG_APPEND)) {
			ret |= (*ctx.func)(fs, &blocks[i],
					    ctx.bcount, 0, i, priv_data);
			if (ret & BLOCK_ABORT)
				goto abort_exit;
		}
	}
	if (*(blocks + EXT2_IND_BLOCK) || (flags & BLOCK_FLAG_APPEND)) {
		ret |= block_iterate_ind(blocks + EXT2_IND_BLOCK,
					 0, EXT2_IND_BLOCK, &ctx);
		if (ret & BLOCK_ABORT)
			goto abort_exit;
	} else
		ctx.bcount += limit;
	if (*(blocks + EXT2_DIND_BLOCK) || (flags & BLOCK_FLAG_APPEND)) {
		ret |= block_iterate_dind(blocks + EXT2_DIND_BLOCK,
					  0, EXT2_DIND_BLOCK, &ctx);
		if (ret & BLOCK_ABORT)
			goto abort_exit;
	} else
		ctx.bcount += limit * limit;
	if (*(blocks + EXT2_TIND_BLOCK) || (flags & BLOCK_FLAG_APPEND)) {
		ret |= block_iterate_tind(blocks + EXT2_TIND_BLOCK,
					  0, EXT2_TIND_BLOCK, &ctx);
		if (ret & BLOCK_ABORT)
			goto abort_exit;
	}

abort_exit:
	if (ret & BLOCK_CHANGED) {
		for (i=0; i < EXT2_N_BLOCKS; i++)
			inode.i_block[i] = blocks[i];
		retval = ext2fs_write_inode(fs, ino, &inode);
		if (retval)
			return retval;
	}

	if (!block_buf)
		ext2fs_free_mem(&ctx.ind_buf);

	return (ret & BLOCK_ERROR) ? ctx.errcode : 0;
}

/*
 * Emulate the old ext2fs_block_iterate function!
 */

struct xlate {
	int (*func)(ext2_filsys	fs,
		    blk_t	*blocknr,
		    int		bcount,
		    void	*priv_data);
	void *real_private;
};

#ifdef __TURBOC__
 #pragma argsused
#endif
static int xlate_func(ext2_filsys fs, blk_t *blocknr, e2_blkcnt_t blockcnt,
		      blk_t ref_block EXT2FS_ATTR((unused)),
		      int ref_offset EXT2FS_ATTR((unused)),
		      void *priv_data)
{
	struct xlate *xl = (struct xlate *) priv_data;

	return (*xl->func)(fs, blocknr, (int) blockcnt, xl->real_private);
}

errcode_t ext2fs_block_iterate(ext2_filsys fs,
			       ext2_ino_t ino,
			       int	flags,
			       char *block_buf,
			       int (*func)(ext2_filsys fs,
					   blk_t	*blocknr,
					   int	blockcnt,
					   void	*priv_data),
			       void *priv_data)
{
	struct xlate xl;
	
	xl.real_private = priv_data;
	xl.func = func;

	return ext2fs_block_iterate2(fs, ino, BLOCK_FLAG_NO_LARGE | flags,
				     block_buf, xlate_func, &xl);
}

#ifdef EXT_DEBUG
void show_header(struct ext3_extent_header *eh)
{
	printf("header: magic=%x entries=%u max=%u depth=%u generation=%u\n",
	       eh->eh_magic, eh->eh_entries, eh->eh_max, eh->eh_depth,
	       eh->eh_generation);
}

void show_index(struct ext3_extent_idx *ix)
{
	printf("index: block=%u leaf=%u leaf_hi=%u unused=%u\n",
	       ix->ei_block, ix->ei_leaf, ix->ei_leaf_hi, ix->ei_unused);
}

void show_extent(struct ext3_extent *ex)
{
	printf("extent: block=%u-%u len=%u start=%u start_hi=%u\n",
	       ex->ee_block, ex->ee_block + ex->ee_len - 1,
	       ex->ee_len, ex->ee_start, ex->ee_start_hi);
}
#else
#define show_header(eh) do { } while (0)
#define show_index(ix) do { } while (0)
#define show_extent(ex) do { } while (0)
#endif

errcode_t ext2fs_extent_header_verify(struct ext3_extent_header *eh, int size)
{
	int eh_max, entry_size;

	show_header(eh);
	if (eh->eh_magic != EXT3_EXT_MAGIC)
		return EXT2_ET_EXTENT_HEADER_BAD;
	if (eh->eh_entries > eh->eh_max)
		return EXT2_ET_EXTENT_HEADER_BAD;
	if (eh->eh_depth == 0)
		entry_size = sizeof(struct ext3_extent);
	else
		entry_size = sizeof(struct ext3_extent_idx);

	eh_max = (size - sizeof(*eh)) / entry_size;
	if (eh->eh_max > eh_max || eh->eh_max < eh_max - 2)
		return EXT2_ET_EXTENT_HEADER_BAD;

	return 0;
}

errcode_t ext2fs_extent_verify(ext2_filsys fs, struct ext3_extent *ex,
			       struct ext3_extent *ex_prev,
			       struct ext3_extent_idx *ix, int ix_len)
{
	show_extent(ex);
	/* FIXME: 48-bit support */
	if (ex->ee_start > fs->super->s_blocks_count)
		return EXT2_ET_EXTENT_LEAF_BAD;

	if (ex->ee_len == 0)
		return EXT2_ET_EXTENT_LEAF_BAD;

	if (ex_prev) {
		/* We can't have a zero logical block except for first index */
		if (ex->ee_block == 0)
			return EXT2_ET_EXTENT_LEAF_BAD;

		/* FIXME: 48-bit support */
		/* extents must be in logical offset order */
		if (ex->ee_block < ex_prev->ee_block + ex_prev->ee_len)
			return EXT2_ET_EXTENT_LEAF_BAD;

		/* extents must not overlap physical blocks */
		if ((ex->ee_start < ex_prev->ee_start + ex_prev->ee_len) &&
		    (ex->ee_start + ex->ee_len > ex_prev->ee_start))
			return EXT2_ET_EXTENT_LEAF_BAD;
	}

	if (ix) {
		/* FIXME: 48-bit support */
		if (ex->ee_block < ix->ei_block)
			return EXT2_ET_EXTENT_LEAF_BAD;

		if (ix_len && ex->ee_block + ex->ee_len > ix->ei_block + ix_len)
			return EXT2_ET_EXTENT_LEAF_BAD;
	}

	return 0;
}

errcode_t ext2fs_extent_index_verify(ext2_filsys fs, struct ext3_extent_idx *ix,
				     struct ext3_extent_idx *ix_prev)
{
	show_index(ix);
	/* FIXME: 48-bit support */
	if (ix->ei_leaf > fs->super->s_blocks_count)
		return EXT2_ET_EXTENT_INDEX_BAD;

	if (ix_prev == NULL)
		return 0;

	/* We can't have a zero logical block except for first index */
	if (ix->ei_block == 0)
		return EXT2_ET_EXTENT_INDEX_BAD;

	if (ix->ei_block <= ix_prev->ei_block)
		return EXT2_ET_EXTENT_INDEX_BAD;

	return 0;
}

errcode_t ext2fs_extent_remove(struct ext3_extent_header *eh,
			       struct ext3_extent *ex)
{
	int entry = ex - EXT_FIRST_EXTENT(eh);

	if (entry < 0 || entry > eh->eh_entries)
		return EXT2_ET_EXTENT_LEAF_BAD;

	memmove(ex, ex + 1, (eh->eh_entries - entry - 1) * sizeof(*ex));
	--eh->eh_entries;

	return 0;
}

errcode_t ext2fs_extent_split(struct ext3_extent_header *eh,
			      struct ext3_extent *ex, int count)
{
	int entry = ex - EXT_FIRST_EXTENT(eh);
	struct ext3_extent *ex_new = ex + 1;

	if (entry < 0 || entry > eh->eh_entries)
		return EXT2_ET_EXTENT_LEAF_BAD;

	if (eh->eh_entries >= eh->eh_max)
		return EXT2_ET_EXTENT_NO_SPACE;

	if (count > ex->ee_len)
		return EXT2_ET_EXTENT_LEAF_BAD;

	if (count > ex->ee_len)
		return EXT2_ET_EXTENT_LEAF_BAD;

	memmove(ex_new, ex, (eh->eh_entries - entry) * sizeof(*ex));
	++eh->eh_entries;

	ex->ee_len = count;
	/* FIXME: 48-bit support */
	ex_new->ee_len -= count;
	ex_new->ee_block += count;
	ex_new->ee_start += count;

	return 0;
}

errcode_t ext2fs_extent_index_remove(struct ext3_extent_header *eh,
				     struct ext3_extent_idx *ix)
{
	struct ext3_extent_idx *first = EXT_FIRST_INDEX(eh);
	int count = ix - first;

	memmove(ix, ix + 1, (eh->eh_entries - count - 1) * sizeof(*ix));
	--eh->eh_entries;

	return 0;
}

/* Internal function for ext2fs_block_iterate().
 * See also block_bmap_extents(). */
int block_iterate_extents(struct ext3_extent_header *eh, blk_t ref_block,
			  int ref_offset EXT2FS_ATTR((unused)),
			  struct block_context *ctx)
{
	int ret = 0;
	int i, j, flags;
	blk_t  block_address;
	struct ext3_extent *ex, *ex_prev = NULL;

	if (eh->eh_depth == 0) {
		ex = EXT_FIRST_EXTENT(eh);
		for (i = 0; i < eh->eh_entries; i++, ex++) {
			show_extent(ex);
			for (j = 0; j < ex->ee_len; j++) {
				block_address = ex->ee_start + j;
				flags = (*ctx->func)(ctx->fs, &block_address,
						     (ex->ee_block + j),
						     ref_block, i,
						     ctx->priv_data);
				if (flags & BLOCK_ABORT) {
					ret |= BLOCK_ABORT;
					return ret;
				}
				if (!(flags & BLOCK_CHANGED))
					continue;

#ifdef EXT_DEBUG
				printf("ugh, extent leaf changed: "
				       "block was %u+%u = %u, now %u\n",
				       ex->ee_start, j,
				       ex->ee_start + j, block_address);
#endif

				/* FIXME: 48-bit support */
				if (ex_prev &&
				    block_address ==
				    ex_prev->ee_start + ex_prev->ee_len &&
				    ex->ee_block + j ==
				    ex_prev->ee_block + ex_prev->ee_len) {
					/* can merge block with prev extent */
					ex_prev->ee_len++;

					ex->ee_len--;
					if (ex->ee_len == 0) {
						/* no blocks left in this one */
						ext2fs_extent_remove(eh, ex);
						i--; ex--;
						break;
					} else {
						ex->ee_start++;
						ex->ee_block++;
						j--;
					}
					ret |= BLOCK_CHANGED;

				} else if (ex->ee_len == 1) {
					/* single-block extent is easy -
					 * change extent directly */
					ex->ee_start = block_address;
					ret |= BLOCK_CHANGED;

				} else if (ext2fs_extent_split(eh, ex, j)) {
					/* split for new block failed */
					/* No multi-level split yet */
					ret |= BLOCK_ABORT | BLOCK_ERROR;
					return ret;

				} else if (j > 0 && (ex + 1)->ee_len > 1 &&
					   ext2fs_extent_split(eh, ex + 1, 1)) {
					/* split after new block failed */
					/* No multi-level split yet */
					ret |= BLOCK_ABORT | BLOCK_ERROR;
					return ret;

				} else if (j == 0) {
					if (ex->ee_len != 1) {
						/* this is an internal error */
						ret |= BLOCK_ABORT |BLOCK_ERROR;
						return ret;
					}
					ex->ee_start = block_address;
					ret |= BLOCK_CHANGED;

				} else {
					ex++;
					i++;
					if (ex->ee_len != 1) {
						/* this is an internal error */
						ret |= BLOCK_ABORT |BLOCK_ERROR;
						return ret;
					}
					ex->ee_start = block_address;
					ret |= BLOCK_CHANGED;
				}
			}
			ex_prev = ex;
		}
	} else {
		char *block_buf;
		struct ext3_extent_idx *ix;
		struct ext3_extent_header *nh;

		ret = ext2fs_get_mem(ctx->fs->blocksize, &block_buf);
		if (ret)
			return ret;

		show_header(eh);
		ix = EXT_FIRST_INDEX(eh);
		for (i = 0; i < eh->eh_entries; i++, ix++) {
			show_index(ix);
			if (!(ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
			    !(ctx->flags & BLOCK_FLAG_DATA_ONLY)) {
				ret |= (*ctx->func)(ctx->fs, &ix->ei_leaf,
						    BLOCK_COUNT_IND, ref_block,
						    i, ctx->priv_data);
				if (ret & BLOCK_ABORT)
					goto free_buf;
			}
			ctx->errcode = ext2fs_read_ext_block(ctx->fs,
							     ix->ei_leaf,
							     block_buf);
			if (ctx->errcode) {
				ret |= BLOCK_ERROR;
				goto free_buf;
			}
			nh = (struct ext3_extent_header *)block_buf;
			if (ext2fs_extent_header_verify(nh,ctx->fs->blocksize)){
				ret |= BLOCK_ERROR;
				goto free_buf;
			}
			flags = block_iterate_extents(nh, ix->ei_leaf, i, ctx);
			if (flags & BLOCK_CHANGED)
				ctx->errcode =
					ext2fs_write_ext_block(ctx->fs,
							       ix->ei_leaf,
							       block_buf);
			if (flags & BLOCK_ABORT) {
				ret |= BLOCK_ABORT;
				goto free_buf;
			}
			if ((ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
			    !(ctx->flags & BLOCK_FLAG_DATA_ONLY))
				ret |= (*ctx->func)(ctx->fs, &ix->ei_leaf,
						     BLOCK_COUNT_IND, ref_block,
						     i, ctx->priv_data);
		}

	free_buf:
		ext2fs_free_mem(&block_buf);
	}
	return ret;
}
