/*
 * extent.c --- iterate over all extents in an inode
 *
 * Copyright (C) 2006 Theodore Ts'o.
 *
 * %Begin-Header%
 * This file may be redistributed under the terms of the GNU Public
 * License.
 * %End-Header%
 */

#include <stdio.h>
#include <string.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <errno.h>

#include "ext2_fs.h"
#include "ext2fs.h"

struct block_context {
	ext2_filsys	fs;
	int (*func)(ext2_filsys	fs,
		    blk_t	*blocknr,
		    e2_blkcnt_t	bcount,
		    blk_t	ref_blk,
		    int		ref_offset,
		    void	*priv_data);
	e2_blkcnt_t	bcount;
	int		bsize;
	int		flags;
	errcode_t	errcode;
	char	*ind_buf;
	char	*dind_buf;
	char	*tind_buf;
	void	*priv_data;
};


#ifdef EXT_DEBUG
void show_header(struct ext3_extent_header *eh)
{
	printf("header: magic=%x entries=%u max=%u depth=%u generation=%u\n",
	       eh->eh_magic, eh->eh_entries, eh->eh_max, eh->eh_depth,
	       eh->eh_generation);
}

void show_index(struct ext3_extent_idx *ix)
{
	printf("index: block=%u leaf=%u leaf_hi=%u unused=%u\n",
	       ix->ei_block, ix->ei_leaf, ix->ei_leaf_hi, ix->ei_unused);
}

void show_extent(struct ext3_extent *ex)
{
	printf("extent: block=%u-%u len=%u start=%u start_hi=%u\n",
	       ex->ee_block, ex->ee_block + ex->ee_len - 1,
	       ex->ee_len, ex->ee_start, ex->ee_start_hi);
}
#else
#define show_header(eh) do { } while (0)
#define show_index(ix) do { } while (0)
#define show_extent(ex) do { } while (0)
#endif

static errcode_t extent_header_verify(struct ext3_extent_header *eh, int size)
{
	int eh_max, entry_size;

	show_header(eh);
	if (eh->eh_magic != EXT3_EXT_MAGIC)
		return EXT2_ET_EXTENT_HEADER_BAD;
	if (eh->eh_entries > eh->eh_max)
		return EXT2_ET_EXTENT_HEADER_BAD;
	if (eh->eh_depth == 0)
		entry_size = sizeof(struct ext3_extent);
	else
		entry_size = sizeof(struct ext3_extent_idx);

	eh_max = (size - sizeof(*eh)) / entry_size;
	if (eh->eh_max > eh_max || eh->eh_max < eh_max - 2)
		return EXT2_ET_EXTENT_HEADER_BAD;

	return 0;
}

static errcode_t extent_verify(ext2_filsys fs, struct ext3_extent *ex,
			       struct ext3_extent *ex_prev,
			       struct ext3_extent_idx *ix, int ix_len)
{
	show_extent(ex);
	/* FIXME: 48-bit support */
	if (ex->ee_start > fs->super->s_blocks_count)
		return EXT2_ET_EXTENT_LEAF_BAD;

	if (ex->ee_len == 0)
		return EXT2_ET_EXTENT_LEAF_BAD;

	if (ex_prev) {
		/* We can't have a zero logical block except for first index */
		if (ex->ee_block == 0)
			return EXT2_ET_EXTENT_LEAF_BAD;

		/* FIXME: 48-bit support */
		/* extents must be in logical offset order */
		if (ex->ee_block < ex_prev->ee_block + ex_prev->ee_len)
			return EXT2_ET_EXTENT_LEAF_BAD;

		/* extents must not overlap physical blocks */
		if ((ex->ee_start < ex_prev->ee_start + ex_prev->ee_len) &&
		    (ex->ee_start + ex->ee_len > ex_prev->ee_start))
			return EXT2_ET_EXTENT_LEAF_BAD;
	}

	if (ix) {
		/* FIXME: 48-bit support */
		if (ex->ee_block < ix->ei_block)
			return EXT2_ET_EXTENT_LEAF_BAD;

		if (ix_len && ex->ee_block + ex->ee_len > ix->ei_block + ix_len)
			return EXT2_ET_EXTENT_LEAF_BAD;
	}

	return 0;
}

static errcode_t extent_index_verify(ext2_filsys fs, struct ext3_extent_idx *ix,
				     struct ext3_extent_idx *ix_prev)
{
	show_index(ix);
	/* FIXME: 48-bit support */
	if (ix->ei_leaf > fs->super->s_blocks_count)
		return EXT2_ET_EXTENT_INDEX_BAD;

	if (ix_prev == NULL)
		return 0;

	/* We can't have a zero logical block except for first index */
	if (ix->ei_block == 0)
		return EXT2_ET_EXTENT_INDEX_BAD;

	if (ix->ei_block <= ix_prev->ei_block)
		return EXT2_ET_EXTENT_INDEX_BAD;

	return 0;
}

static errcode_t extent_remove(struct ext3_extent_header *eh,
			       struct ext3_extent *ex)
{
	int entry = ex - EXT_FIRST_EXTENT(eh);

	if (entry < 0 || entry > eh->eh_entries)
		return EXT2_ET_EXTENT_LEAF_BAD;

	memmove(ex, ex + 1, (eh->eh_entries - entry - 1) * sizeof(*ex));
	--eh->eh_entries;

	return 0;
}

static errcode_t extent_split(struct ext3_extent_header *eh,
			      struct ext3_extent *ex, int count)
{
	int entry = ex - EXT_FIRST_EXTENT(eh);
	struct ext3_extent *ex_new = ex + 1;

	if (entry < 0 || entry > eh->eh_entries)
		return EXT2_ET_EXTENT_LEAF_BAD;

	if (eh->eh_entries >= eh->eh_max)
		return EXT2_ET_EXTENT_NO_SPACE;

	if (count > ex->ee_len)
		return EXT2_ET_EXTENT_LEAF_BAD;

	if (count > ex->ee_len)
		return EXT2_ET_EXTENT_LEAF_BAD;

	memmove(ex_new, ex, (eh->eh_entries - entry) * sizeof(*ex));
	++eh->eh_entries;

	ex->ee_len = count;
	/* FIXME: 48-bit support */
	ex_new->ee_len -= count;
	ex_new->ee_block += count;
	ex_new->ee_start += count;

	return 0;
}

static errcode_t extent_index_remove(struct ext3_extent_header *eh,
				     struct ext3_extent_idx *ix)
{
	struct ext3_extent_idx *first = EXT_FIRST_INDEX(eh);
	int count = ix - first;

	memmove(ix, ix + 1, (eh->eh_entries - count - 1) * sizeof(*ix));
	--eh->eh_entries;

	return 0;
}

/* Internal function for ext2fs_block_iterate().
 * See also block_bmap_extents(). */
static int block_iterate_extents(struct ext3_extent_header *eh, blk_t ref_block,
			  int ref_offset EXT2FS_ATTR((unused)),
			  struct block_context *ctx)
{
	int ret = 0;
	int i, j, flags;
	blk_t  block_address;
	struct ext3_extent *ex, *ex_prev = NULL;

	if (eh->eh_depth == 0) {
		ex = EXT_FIRST_EXTENT(eh);
		for (i = 0; i < eh->eh_entries; i++, ex++) {
			show_extent(ex);
			for (j = 0; j < ex->ee_len; j++) {
				block_address = ex->ee_start + j;
				flags = (*ctx->func)(ctx->fs, &block_address,
						     (ex->ee_block + j),
						     ref_block, i,
						     ctx->priv_data);
				if (flags & BLOCK_ABORT) {
					ret |= BLOCK_ABORT;
					return ret;
				}
				if (!(flags & BLOCK_CHANGED))
					continue;

#ifdef EXT_DEBUG
				printf("ugh, extent leaf changed: "
				       "block was %u+%u = %u, now %u\n",
				       ex->ee_start, j,
				       ex->ee_start + j, block_address);
#endif

				/* FIXME: 48-bit support */
				if (ex_prev &&
				    block_address ==
				    ex_prev->ee_start + ex_prev->ee_len &&
				    ex->ee_block + j ==
				    ex_prev->ee_block + ex_prev->ee_len) {
					/* can merge block with prev extent */
					ex_prev->ee_len++;

					ex->ee_len--;
					if (ex->ee_len == 0) {
						/* no blocks left in this one */
						extent_remove(eh, ex);
						i--; ex--;
						break;
					} else {
						ex->ee_start++;
						ex->ee_block++;
						j--;
					}
					ret |= BLOCK_CHANGED;

				} else if (ex->ee_len == 1) {
					/* single-block extent is easy -
					 * change extent directly */
					ex->ee_start = block_address;
					ret |= BLOCK_CHANGED;

				} else if (extent_split(eh, ex, j)) {
					/* split for new block failed */
					/* No multi-level split yet */
					ret |= BLOCK_ABORT | BLOCK_ERROR;
					return ret;

				} else if (j > 0 && (ex + 1)->ee_len > 1 &&
					   extent_split(eh, ex + 1, 1)) {
					/* split after new block failed */
					/* No multi-level split yet */
					ret |= BLOCK_ABORT | BLOCK_ERROR;
					return ret;

				} else if (j == 0) {
					if (ex->ee_len != 1) {
						/* this is an internal error */
						ret |= BLOCK_ABORT |BLOCK_ERROR;
						return ret;
					}
					ex->ee_start = block_address;
					ret |= BLOCK_CHANGED;

				} else {
					ex++;
					i++;
					if (ex->ee_len != 1) {
						/* this is an internal error */
						ret |= BLOCK_ABORT |BLOCK_ERROR;
						return ret;
					}
					ex->ee_start = block_address;
					ret |= BLOCK_CHANGED;
				}
			}
			ex_prev = ex;
		}
	} else {
		char *block_buf;
		struct ext3_extent_idx *ix;
		struct ext3_extent_header *nh;

		ret = ext2fs_get_mem(ctx->fs->blocksize, &block_buf);
		if (ret)
			return ret;

		show_header(eh);
		ix = EXT_FIRST_INDEX(eh);
		for (i = 0; i < eh->eh_entries; i++, ix++) {
			show_index(ix);
			if (!(ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
			    !(ctx->flags & BLOCK_FLAG_DATA_ONLY)) {
				ret |= (*ctx->func)(ctx->fs, &ix->ei_leaf,
						    BLOCK_COUNT_IND, ref_block,
						    i, ctx->priv_data);
				if (ret & BLOCK_ABORT)
					goto free_buf;
			}
			ctx->errcode = ext2fs_read_ext_block(ctx->fs,
							     ix->ei_leaf,
							     block_buf);
			if (ctx->errcode) {
				ret |= BLOCK_ERROR;
				goto free_buf;
			}
			nh = (struct ext3_extent_header *)block_buf;
			if (extent_header_verify(nh,ctx->fs->blocksize)){
				ret |= BLOCK_ERROR;
				goto free_buf;
			}
			flags = block_iterate_extents(nh, ix->ei_leaf, i, ctx);
			if (flags & BLOCK_CHANGED)
				ctx->errcode =
					ext2fs_write_ext_block(ctx->fs,
							       ix->ei_leaf,
							       block_buf);
			if (flags & BLOCK_ABORT) {
				ret |= BLOCK_ABORT;
				goto free_buf;
			}
			if ((ctx->flags & BLOCK_FLAG_DEPTH_TRAVERSE) &&
			    !(ctx->flags & BLOCK_FLAG_DATA_ONLY))
				ret |= (*ctx->func)(ctx->fs, &ix->ei_leaf,
						     BLOCK_COUNT_IND, ref_block,
						     i, ctx->priv_data);
		}

	free_buf:
		ext2fs_free_mem(&block_buf);
	}
	return ret;
}


errcode_t ext2fs_extent_iterate2(ext2_filsys fs,
				 ext2_ino_t ino,
				 int	flags,
				 char *block_buf,
				 int (*func)(ext2_filsys fs,
					     struct ext2fs_extent *extent,
					     void	*priv_data),
				 int (*meta_func)(ext2_filsys fs,
						  blk64_t blk,
						  int blk_type,
						  char *buf,
						  void *priv_data),
				 void *priv_data)
{
	struct ext2_inode inode;
	errcode_t errcode;
	struct ext3_extent_header *eh;

	EXT2_CHECK_MAGIC(fs, EXT2_ET_MAGIC_EXT2FS_FILSYS);

	errcode = ext2fs_read_inode(fs, ino, &inode);
	if (errcode)
		return errcode;

	if (!(inode.i_flags & EXT4_EXTENTS_FL)) {
		/*
		 * XXX not yet....
		 */
		return -EOPNOTSUPP;
	}

	eh = (struct ext3_extent_header *) inode.i_blocks;

}
